/*
    Project: TCGINA
    Version: V1.18
    File   : tcpart.c

    Copyright (c) 2005-2008 Author of TCGINA DLL. All rights reserved.
    Public Key Fingerprint: 294B A769 4A0A CC05 DAE6 00DD FF47 8C72 4097 67CE

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCGINA binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include "tcpriv.h"
#include <tchar.h>
#include <stdio.h>

/*------------------------------------------------------------------------------
// Partition data
//
// This implementation relies on following two assumptions:
// (1) For each device "\Device\Harddisk<n>\Partition0" exists a symbolic link
//     with the name "PhysicalDrive<n>" where <n> is the number of the hard
//     disk starting with zero.
// (2) All partitions of hard disk number <n> are accessable as
//     "\Device\Harddisk<n>\Partition<m>" for <m>=1,2,...,M
//     where M is the number of partitions.
//
// The latter assumption can fail if a partition has been removed and the
// corresponding hard disk has not been remounted.
//
// Alternative implementations are:
// (1) Iterate through all possible disk and partition numbers and let the
//     mount procedure fail if the actual partition is not available.
// (2) Use NTDLL.DLL to enumerate the "\Device\Harddisk<n>\Partition<m>"
//     device names.
//                                                                            */

#define TC_MAX_DISK 63
#define TC_MAX_PARTITION 32

static BOOL WINAPI TC_GetNumberOfPartitions(PTC_PARTITION_DATA pd)
{
    HANDLE hDrive;
    TCHAR szSymbolicLink[17+2+1];

    TC_Assert(pd);
    TC_Assert(pd->dwDisk <= TC_MAX_DISK);

    if (_sntprintf(szSymbolicLink, countof(szSymbolicLink),
      TEXT("\\\\.\\PhysicalDrive%d"), pd->dwDisk) <= 0) {
        pd->dwPartitionCount = 0;
        pd->dwPartition = (DWORD) -1;
        return FALSE;
    }

    hDrive = CreateFile(szSymbolicLink, GENERIC_READ,
      FILE_SHARE_READ|FILE_SHARE_WRITE, 0, OPEN_EXISTING, 0, NULL);

    if (hDrive != INVALID_HANDLE_VALUE) {

        BYTE cbOutBuffer[sizeof(DRIVE_LAYOUT_INFORMATION)+
          TC_MAX_PARTITION * sizeof(PARTITION_INFORMATION)];
        DWORD nBytesReturned;

        if (DeviceIoControl(hDrive, IOCTL_DISK_GET_DRIVE_LAYOUT, NULL, 0,
          &cbOutBuffer, sizeof(cbOutBuffer), &nBytesReturned, 0)) {
            DRIVE_LAYOUT_INFORMATION *dli =
              (DRIVE_LAYOUT_INFORMATION*)&cbOutBuffer;
            PARTITION_INFORMATION *pinfo = dli->PartitionEntry;
            PARTITION_INFORMATION *pinfo_end = pinfo + dli->PartitionCount;
            pd->dwPartitionCount = 0;
            for (; pinfo != pinfo_end; ++pinfo) {
                if (pinfo->RecognizedPartition) pd->dwPartitionCount++;
            }
        } else {
            switch(GetLastError()) {
            case ERROR_NOT_READY:
                pd->dwPartitionCount = 1;
                break;
            case ERROR_INSUFFICIENT_BUFFER:
            default:
                pd->dwPartitionCount = TC_MAX_PARTITION + 1;
            }
        }
        CloseHandle(hDrive);

    } else {

        if (GetLastError() == ERROR_ACCESS_DENIED) {
            pd->dwPartitionCount = TC_MAX_PARTITION + 1;
        } else {
            pd->dwPartitionCount = 0;
            pd->dwPartition = (DWORD) -1;
            return FALSE;
        }

    }

    pd->dwPartition = 0;

    return TRUE;
}

BOOL WINAPI TC_GetFirstPartition(PTC_PARTITION_DATA pd)
{
    TC_Assert(pd);

    TC_ZeroMemory(pd, sizeof(*pd));

    do {
        if (TC_GetNumberOfPartitions(pd)) return TRUE;
    } while (++pd->dwDisk <= TC_MAX_DISK);

    return FALSE;
}

BOOL WINAPI TC_GetNextPartition(PTC_PARTITION_DATA pd)
{
    TC_Assert(pd);

    if (pd->dwDisk > TC_MAX_DISK) return FALSE;
    if (pd->dwPartition < pd->dwPartitionCount) {
        ++pd->dwPartition;
        return TRUE;
    }
    while (++pd->dwDisk <= TC_MAX_DISK) {
        if (TC_GetNumberOfPartitions(pd)) return TRUE;
    }

    return FALSE;
}

BOOL WINAPI TC_GetPartitionDeviceName(const PTC_PARTITION_DATA pd,
  LPTSTR pszDest, DWORD nSize)
{
    TC_Assert(pd);
    TC_Assert(pszDest);

    if (pd->dwPartition > pd->dwPartitionCount) return FALSE;
    if (pd->dwDisk > TC_MAX_DISK) return FALSE;
    if (pd->dwPartition > TC_MAX_PARTITION) return FALSE;

    return _sntprintf(pszDest, nSize,
      TEXT("\\Device\\Harddisk%d\\Partition%d"),
      pd->dwDisk, pd->dwPartition) > 0;
}

BOOL WINAPI TC_GetPartitionData(PTC_PARTITION_DATA pd,
  LPCTSTR pszPartitionName)
{
    LPTSTR pszTemp;

    TC_Assert(pd);
    TC_Assert(pszPartitionName);

    if (pd == NULL || pszPartitionName == NULL) return FALSE;

    if (_tcsnicmp(pszPartitionName, TEXT("\\Device\\Harddisk"), 16) != 0) {
        return FALSE;
    }
    pszPartitionName += 16;

    pd->dwPartitionCount = 0;
    pd->dwDisk = _tcstoul(pszPartitionName, &pszTemp, 10);
    if (pszTemp == pszPartitionName) return FALSE;
    if (_tcsnicmp(pszTemp, TEXT("\\Partition"), 10) != 0) {
        return FALSE;
    }

    pszPartitionName = pszTemp + 10;
    pd->dwPartition = _tcstoul(pszPartitionName, &pszTemp, 10);
    if (pszTemp == pszPartitionName || *pszTemp != TEXT('\0')) return FALSE;

    return TRUE;
}
